"""
    sphinxcontrib.htmlhelp.version
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    :copyright: Copyright 2007-2019 by the Sphinx team, see README.
    :license: BSD, see LICENSE for details.
"""

__version__ = '2.0.0'
__version_info__ = tuple(map(int, __version__.split('.')))
