# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

try:
    from asyncio import create_task
except ImportError:
    # create_task is not present in Python 3.6. Once Airflow is at 3.7+, we can
    # remove this helper.
    def create_task(*args, **kwargs):  # type: ignore
        """A version of create_task that always errors."""
        raise RuntimeError("Airflow's async functionality is only available on Python 3.7+")


__all__ = ["create_task"]
